package ru.dbotthepony.mc.otm.menu

import net.minecraft.network.chat.Component
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.Items
import ru.dbotthepony.mc.otm.client.render.Widgets18
import ru.dbotthepony.mc.otm.client.render.sprites.AbstractMatterySprite
import ru.dbotthepony.mc.otm.container.IFilteredContainerSlot
import ru.dbotthepony.mc.otm.container.util.containerSlotOrNull
import ru.dbotthepony.mc.otm.core.TranslatableComponent
import ru.dbotthepony.mc.otm.core.isNotEmpty
import ru.dbotthepony.mc.otm.container.ItemStackKey
import ru.dbotthepony.mc.otm.container.asKey
import ru.dbotthepony.mc.otm.container.asKeyOrNull

class QuickMoveInput(private val menu: MatteryMenu, val from: Collection<Slot>, val to: Collection<Slot>, val mode: Mode, val dontTouchFilteredSlots: Boolean = true) {
	/**
	 * slots with items come first
	 */
	object HasItemComparator : Comparator<Slot> {
		override fun compare(a: Slot, b: Slot): Int {
			val hasItemA = a.item.isNotEmpty
			val hasItemB = b.item.isNotEmpty

			return hasItemB.compareTo(hasItemA)
		}
	}

	/**
	 * slots with filters come first
	 */
	object HasFilterComparator : Comparator<Slot> {
		override fun compare(a: Slot, b: Slot): Int {
			val slotA = a.containerSlotOrNull()
			val slotB = b.containerSlotOrNull()

			val hasFilterA = slotA is IFilteredContainerSlot && slotA.filter.hasRules
			val hasFilterB = slotB is IFilteredContainerSlot && slotB.filter.hasRules

			return hasFilterB.compareTo(hasFilterA)
		}
	}

	enum class Mode(val iconFromStorage: AbstractMatterySprite, val iconToStorage: AbstractMatterySprite) {
		RESTOCK(
			Widgets18.RESTOCK_FROM_STORAGE,
			Widgets18.RESTOCK_TO_STORAGE
		) {
			override fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player, dontTouchFilteredSlots: Boolean) {
				if (from.isEmpty() || to.isEmpty()) return
				val (_, itemsFrom) = computeSlotLists(from, dontTouchFilteredSlots)
				val (_, itemsTo, filteredTo) = computeSlotLists(to, false)

				val intersect: Collection<ItemStackKey>

				if (filteredTo.isNotEmpty())
					intersect = itemsFrom.keys
				else if (itemsFrom.size < itemsTo.size)
					intersect = itemsFrom.keys.filter { it in itemsTo.keys }
				else
					intersect = itemsTo.keys.filter { it in itemsFrom.keys }

				for (key in intersect) {
					val slotsTo = ArrayList(itemsTo[key] ?: listOf())
					slotsTo.addAll(0, filteredTo)
					val slotsFrom = itemsFrom[key]!!

					slotsFrom.forEach { moveItemStackTo(player, it, slotsTo) }
				}
			}
		},

		RESTOCK_WITH_MOVE(
			Widgets18.RESTOCK_WITH_MOVE_FROM_STORAGE,
			Widgets18.RESTOCK_WITH_MOVE_TO_STORAGE
		) {
			override fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player, dontTouchFilteredSlots: Boolean) {
				if (from.isEmpty() || to.isEmpty()) return
				val (_, itemsFrom) = computeSlotLists(from, dontTouchFilteredSlots)
				val (emptyTo, itemsTo, filteredTo) = computeSlotLists(to, false)

				val intersect: Collection<ItemStackKey>

				if (filteredTo.isNotEmpty())
					intersect = itemsFrom.keys
				else if (itemsFrom.size < itemsTo.size)
					intersect = itemsFrom.keys.filter { it in itemsTo.keys }
				else
					intersect = itemsTo.keys.filter { it in itemsFrom.keys }

				for (key in intersect) {
					val slotsTo = ArrayList(itemsTo[key] ?: listOf()).also { it.addAll(0, filteredTo) }
					if (slotsTo.isEmpty()) continue
					prioritySortSlotsInPlace(slotsTo, key.asItemStack())

					val slotsFrom = itemsFrom[key]!!

					slotsFrom.removeIf { moveItemStackTo(player, it, slotsTo, sort = false); it.item.isEmpty }
					var moveAny = false
					slotsFrom.forEach { moveAny = moveItemStackTo(player, it, emptyTo, sort = false) || moveAny }
					if (moveAny) emptyTo.removeIf { it.item.isNotEmpty }
				}
			}
		},

		MOVE(
			Widgets18.MOVE_EVERYTHING_FROM_STORAGE,
			Widgets18.MOVE_EVERYTHING_TO_STORAGE
		) {
			override fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player, dontTouchFilteredSlots: Boolean) {
				if (from.isEmpty() || to.isEmpty()) return
				val toSorted = prioritySortSlots(to)

				from.forEach {
					val slot = it.containerSlotOrNull()

					if (!dontTouchFilteredSlots || slot !is IFilteredContainerSlot || !slot.filter.hasRules)
						moveItemStackTo(player, it, toSorted, sort = false)
				}
			}
		};

		abstract fun move(from: Collection<Slot>, to: Collection<Slot>, player: Player, dontTouchFilteredSlots: Boolean = true)

		val textFromStorage: Component get() {
			return TranslatableComponent("otm.gui.quickmove_from.${name.lowercase()}")
		}

		val textToStorage: Component get() {
			return TranslatableComponent("otm.gui.quickmove_to.${name.lowercase()}")
		}
	}

	val input = menu.oneWayInput(handler = ::handle)

	fun clientInput() {
		input.accept(null)
	}

	private fun handle() {
		mode.move(from, to, menu.player, dontTouchFilteredSlots)
	}

	private data class SlotLists(
		val empty: MutableList<Slot>,
		val withItems: MutableMap<ItemStackKey, MutableList<Slot>>,
		val withFilters: MutableList<Slot>,
	)

	companion object {
		fun create(menu: MatteryMenu, from: Collection<Slot>, to: Collection<Slot>, dontTouchFilteredSlots: Boolean = true): Map<Mode, QuickMoveInput> {
			return Mode.entries.associateWith { QuickMoveInput(menu, from, to, it, dontTouchFilteredSlots) }
		}

		private fun computeSlotLists(slots: Collection<Slot>, skipFilteredSlots: Boolean): SlotLists {
			val emptySlots = ArrayList<Slot>()
			val filteredSlots = ArrayList<Slot>()
			val filledSlots = HashMap<ItemStackKey, MutableList<Slot>>()

			for (slot in slots) {
				val underlyingSlot = slot.containerSlotOrNull()

				if (underlyingSlot is IFilteredContainerSlot && (underlyingSlot.filter.denyAll || !underlyingSlot.filter.allowAll && skipFilteredSlots))
					continue

				val key = slot.item.asKeyOrNull()

				if (key == null) {
					if (underlyingSlot is IFilteredContainerSlot && underlyingSlot.filter.hasRules) {
						filteredSlots.add(slot)
					} else {
						emptySlots.add(slot)
					}
				} else {
					filledSlots.computeIfAbsent(key) { ArrayList() }.add(slot)
				}
			}

			return SlotLists(emptySlots, filledSlots, filteredSlots)
		}

		fun moveItemStackTo(
			player: Player,
			source: Slot,
			slots: Collection<Slot>,
			sort: Boolean = true
		): Boolean {
			if (!source.mayPickup(player) || source.item.isEmpty || slots.isEmpty())
				return false

			val remainder = moveItemStackToSlots(source.item, slots, sort = sort)

			if (remainder.count == source.item.count)
				return false

			val copy = source.item.copy()

			if (remainder.isEmpty) {
				source.setByPlayer(ItemStack.EMPTY)
				source.onTake(player, copy)
			} else {
				copy.count = source.item.count - remainder.count
				source.item.count = remainder.count
				source.onTake(player, copy)
			}

			return true
		}

		private val itemFilterSlotComparator = HasItemComparator.thenComparing(HasFilterComparator)

		fun <T : MutableList<Slot>> prioritySortSlotsInPlace(slots: T, filterItem: ItemStack? = null): T {
			slots.removeIf {
				val slot = it.containerSlotOrNull()

				it.isOverCapacity ||
					filterItem != null && !it.mayPlace(filterItem) ||
					slot is IFilteredContainerSlot && (slot.filter.denyAll || filterItem != null && !slot.filter.test(filterItem))
			}

			slots.sortWith(itemFilterSlotComparator)
			return slots
		}

		fun prioritySortSlots(slots: Collection<Slot>, filterItem: ItemStack? = null): MutableList<Slot> {
			return prioritySortSlotsInPlace(ArrayList(slots), filterItem)
		}

		fun moveItemStackToSlots(item: ItemStack, slots: Collection<Slot>, simulate: Boolean = false, sort: Boolean = true): ItemStack {
			if (item.isEmpty)
				return ItemStack.EMPTY
			else if (slots.isEmpty())
				return item.copy()

			val sortedSlots = if (sort) prioritySortSlots(slots, item) else slots
			val copy = item.copy()

			for (slot in sortedSlots) {
				val limit = slot.getMaxStackSize(copy)

				if (!slot.hasItem()) {
					val newCount = copy.count.coerceAtMost(limit)

					if (!simulate) {
						slot.setByPlayer(copy.copy().also { it.count = newCount })
						// slot.setChanged()
					}

					copy.shrink(newCount)

					if (copy.isEmpty)
						return ItemStack.EMPTY
				} else if (limit > slot.item.count && ItemStack.isSameItemSameComponents(slot.item, copy)) {
					val newCount = (slot.item.count + copy.count).coerceAtMost(limit)
					val diff = newCount - slot.item.count
					copy.count -= diff

					if (!simulate) {
						slot.item.count += diff
						slot.setChanged()
					}

					if (copy.isEmpty)
						return ItemStack.EMPTY
				}
			}

			return copy
		}
	}
}
